#!/usr/bin/env python3

from argparse import ArgumentParser
from pathlib import Path
import os
import re
import subprocess
import sys
import tempfile

parser = ArgumentParser()

parser.add_argument("RUSTC", type=Path, help="Path to rustc")
parser.add_argument("--query", action="store",
                    choices=['native-static-libs', 'default-host-toolchain', 'stable-actual-version'],
                    help="Item to query from RustC ['native-static-libs', 'default-host-toolchain', 'stable-actual-version']",
                    required=True)
parser.add_argument("--toolchain-version", action="store",
                    help="Rust Toolchain Version (if needed)")
parser.add_argument("--target", help="Target triplet")
parser.add_argument("--build-triplet", help="Build machine triplet (for cross builds using specific toolchain version)")

def removeprefix_fallback(s, pfx):
    if sys.version_info > (3, 9):
        return s.removeprefix(pfx)
    elif s.startswith(pfx):
        return s[len(pfx):]
    else:
        return s

def removesuffix_fallback(s, sfx):
    if sys.version_info > (3, 9):
        return s.removesuffix(sfx)
    elif s.endswith(sfx):
        return s[:-len(sfx)]
    else:
        return s

def retrieve_native_static_libs_from_output(output):
    if re.match(r'^error[:|\[]', output):
        print(output, file=sys.stderr)
        sys.exit()

    for i in output.strip().splitlines():
        match = re.match(r".+native-static-libs: (.+)", i)
        if match:
            libs = match.group(1).split()
            libs = [removesuffix_fallback(lib, ".lib") for lib in libs] # msvc
            libs = [removeprefix_fallback(lib, "-l") for lib in libs] # msys2
            print(
                " ".join(
                    set(
                        libs
                    )
                )
            )

# Get the default target host or actual version of toolchain
def retrive_version_info(output, query):
    for i in output.strip().splitlines():
        match = re.match(r"%s: (.+)" % query, i)
        if match:
            return match.group(1)

if __name__ == "__main__":
    args = parser.parse_args()
    query = args.query
    query_arg = None
    rustc_cmd = [Path(args.RUSTC).as_posix()]

    if args.toolchain_version is not None:
        if args.target is None and args.build_triplet is None:
            raise ValueError('--target or --build-triplet argument required if --toolchain-version is used')
        if args.build_triplet is not None:
            rustc_cmd.extend(['+%s-%s' % (args.toolchain_version, args.build_triplet)])
        else:
            rustc_cmd.extend(['+%s-%s' % (args.toolchain_version, args.target)])
        
    if query == 'native-static-libs':
        query_arg = ['--print=%s' % query]
    else:
        query_arg = ['--version', '--verbose']
    rustc_cmd.extend(query_arg)
    if args.target:
        rustc_cmd.extend(['--target', args.target])

    fd, dummy_out = tempfile.mkstemp()
    os.close(fd)
    try:
        # We need these for '--print=native-static-libs' on Windows
        if query == 'native-static-libs':
            rustc_cmd.extend(['--crate-type', 'staticlib'])
            rustc_cmd.append(os.devnull)
            rustc_cmd.extend(['-o', dummy_out])

        query_results = subprocess.run(
            rustc_cmd,
            capture_output=True,
            text=True,
        )
    finally:
        os.unlink(dummy_out)

    if query == 'native-static-libs':
        retrieve_native_static_libs_from_output(query_results.stderr)
    elif query == 'default-host-toolchain' or query == 'stable-actual-version':
        if query_results.stderr == '':
            if query == 'default-host-toolchain':
                result = retrive_version_info(query_results.stdout, 'host')
            else:
                result = retrive_version_info(query_results.stdout, 'release')
            print(result)
        else:
            print(query_results.stderr, file=sys.stderr)



# Checks that the example Cargo.toml snippet from rsvg/src/lib.rs has the same versions for
# dependencies that librsvg uses during compilation.

import sys
import toml

# Looks for a crate version in the 'dependencies' section of a TOML document, either of these:
#
# [dependencies]
# foo = "1.2.3"
# bar = { version = "4.5.6", features=["something", "else", "here"]
def get_crate_version(toml_doc, crate_name):
    if 'dependencies' in toml_doc:
        crate_decl = toml_doc['dependencies'][crate_name]
    else:
        crate_decl = toml_doc['workspace']['dependencies'][crate_name]

    if isinstance(crate_decl, str):
        version = crate_decl
    else:
        version = crate_decl['version']

    return version

# Given a Rust file that has a toplevel comment somewhere like
#
#   //! ```toml
#   //! [dependencies]
#   //! librsvg = "2.57.0-beta.2"
#   //! cairo-rs = "0.18"
#   //! gio = "0.18"   # only if you need streams
#   //! ```
#
# extracts just the TOML as a string, without the //! prefix.
def find_toml_in_rust_toplevel_docs(lines):
    found_start = False
    start_index = 0
    end_index = 0

    for (i, line) in enumerate(lines):
        if not found_start and line.startswith('//! ```toml'):
            found_start = True
            start_index = i
            end_index = i
        elif found_start and line.startswith('//! ```'):
            end_index = i
            break

    if not found_start:
        raise Exception(
            "did not find start of ```toml block in the toplevel documentation comments"
        )

    snippet = lines[(start_index + 1):end_index]

    without_comment = [s.removeprefix('//! ') for s in snippet]

    return "".join(without_comment)

def check_dependency_version(cargo_toml_filename, cargo_toml, other_filename, other_toml,
                             dependency_name):
    dep_in_cargo_toml = get_crate_version(cargo_toml, dependency_name)
    dep_in_other = get_crate_version(other_toml, dependency_name)

    if dep_in_cargo_toml != dep_in_other:
        raise Exception(
            f"""{dependency_name} version in {cargo_toml_filename} is {dep_in_cargo_toml} but 
            is referenced in {other_filename} as {dep_in_other}"""
        )

def check():
    cargo_toml = toml.load('rsvg/Cargo.toml')
    librsvg_version = cargo_toml['package']['version']

    example_file = open('rsvg/src/lib.rs')
    example_contents = example_file.readlines()
    example_toml_str = find_toml_in_rust_toplevel_docs(example_contents)
    example_toml = toml.loads(example_toml_str)

    example_version = get_crate_version(example_toml, 'librsvg')

    if librsvg_version != example_version:
        raise Exception(
            f"""librsvg version in rsvg/Cargo.toml is {librsvg_version} but is referenced as
            {example_version} in rsvg/src/lib.rs"""
        )

    DEPENDENCIES = [
        'cairo-rs',
        'gio',
    ]

    cargo_toml = toml.load('Cargo.toml')
    for dependency_name in DEPENDENCIES:
        check_dependency_version(
            'Cargo.toml',
            cargo_toml,
            'rsvg/src/lib.rs',
            example_toml,
            dependency_name
        )

    print("Dependency versions match in rsvg/src/lib.rs.  All good!", file=sys.stderr)

if __name__ == '__main__':
    check()



# This script checks that the Minimum Supported Rust Version (MSRV) has the same value
# in several places throughout the source tree.

import re
import sys

PLACES_WITH_RUST_VERSION = [
    ['meson.build',                      r"msrv = '(.*)'"],
    ['Cargo.toml',                       r'rust-version\s*=\s*"(.*)"'],
    ['ci/container_builds.yml',          r'RUST_MINIMUM:\s*"(.*)"'],
    ['devel-docs/_build_dependencies.rst', r'`rust .*`_ (.*) or later'],
]

PLACES_WITH_CARGO_CBUILD_VERSION = [
    ['meson.build',           r"cargo_cbuild_version = '(.*)'"],
    ['librsvg-c/Cargo.toml',  r'min_version = "(.*)"'],
]

def check_versions(name, places):
    versions = []

    for filename, regex in places:
        r = re.compile(regex)

        with open(filename) as f:
            matched = False
            for idx, line in enumerate(f.readlines()):
                matches = r.search(line)
                if matches is not None:
                    matched = True
                    line_number = idx + 1
                    versions.append([filename, line_number, matches.group(1), line])

            if not matched:
                raise Exception(f'file {filename} does not have a line that matches {regex}')

    assert len(versions) > 0

    all_the_same = True

    for filename, line_number, version, line in versions[1:]:
        if version != versions[0][2]:
            all_the_same = False

    if not all_the_same:
        print(f'{name}: Version numbers do not match in these lines, please fix them!\n', file=sys.stderr)

        for filename, line_number, version, line in versions:
            print(f'  {filename}:{line_number}: {line}', file=sys.stderr)

        sys.exit(1)

    print(f'{name}: Versions number match.  All good!', file=sys.stderr)

def main():
    check_versions('rustc', PLACES_WITH_RUST_VERSION)
    check_versions('cargo-cbuild', PLACES_WITH_CARGO_CBUILD_VERSION)

if __name__ == "__main__":
    main()



"""Custom reStructuredText entities/helpers for referencing entities in the
librsvg internals documentation.

For any changes made in this module, please ensure
``devel-docs/devel_docs_mod_guide.rst`` is updated accordingly, if necessary.

For help/reference on modifying these, see:

- https://www.sphinx-doc.org/en/master/development/tutorials/extending_syntax.html
- https://docutils.sourceforge.io/docs/howto/rst-roles.html
- https://protips.readthedocs.io/link-roles.html
- https://www.sphinx-doc.org/en/master/development/tutorials/adding_domain.html
- https://www.sphinx-doc.org/en/master/extdev/utils.html#sphinx.util.docutils.SphinxRole
- https://www.sphinx-doc.org/en/master/extdev/utils.html#sphinx.util.docutils.ReferenceRole
- https://github.com/sphinx-doc/sphinx/blob/master/sphinx/roles.py
"""

from __future__ import annotations

from docutils.nodes import literal, reference
from sphinx.domains import Domain
from sphinx.util.docutils import ReferenceRole

BASE_URL = "https://gnome.pages.gitlab.gnome.org/librsvg/internals"


class CrateRole(ReferenceRole):
    """A custom Sphinx role for referencing crates in the librsvg internals
    documentation.
    """

    def run(self):
        node = reference(
            self.rawtext,
            refuri=f"{BASE_URL}/{self.target}/index.html",
            **self.options,
        )
        node += literal(self.rawtext, self.title)

        return [node], []


class ModuleRole(ReferenceRole):
    """A custom Sphinx role for referencing modules in the librsvg internals
    documentation.
    """

    def run(self):
        components = self.target.split("::")

        try:
            *parents, module = components
            if not parents:
                raise ValueError
        except ValueError:
            msg = self.inliner.reporter.error(
                f"Invalid module target: {self.target!r}", line=self.lineno
            )
            prb = self.inliner.problematic(self.rawtext, self.rawtext, msg)

            return [prb], [msg]

        node = reference(
            self.rawtext,
            refuri=f"{BASE_URL}/{'/'.join(parents)}/{module}/index.html",
            **self.options,
        )
        node += literal(
            self.rawtext, self.title if self.has_explicit_title else module
        )

        return [node], []


class TopLevelRole(ReferenceRole):
    """A custom Sphinx role for referencing top-level entities in the librsvg
    internals documentation.
    """

    def __init__(self, target_kind: str, *, target_is_callable: bool = False):
        super().__init__()
        self.target_kind = target_kind
        self.target_is_callable = target_is_callable

    def run(self):
        *parents, item = self.target.split("::")

        if not parents:
            msg = self.inliner.reporter.error(
                f"Invalid {self.target_kind} target: {self.target!r}",
                line=self.lineno,
            )
            prb = self.inliner.problematic(self.rawtext, self.rawtext, msg)

            return [prb], [msg]

        node = reference(
            self.rawtext,
            refuri=(
                f"{BASE_URL}/{'/'.join(parents)}/{self.target_kind}"
                f".{item}.html"
            ),
            **self.options,
        )
        node += literal(
            self.rawtext,
            (
                self.title if self.has_explicit_title
                else f"{item}()" if self.target_is_callable
                else item
            ),
        )

        return [node], []


class MemberRole(ReferenceRole):
    """A custom Sphinx role for referencing members of structs, enums, etc
    in the librsvg internals documentation.
    """

    def __init__(
        self,
        target_kind: str,
        parent_tag: str,
        member_tag: str,
        *,
        target_is_callable: bool = False,
    ):
        super().__init__()
        self.target_kind = target_kind
        self.parent_tag = parent_tag
        self.member_tag = member_tag
        self.target_is_callable = target_is_callable

    def run(self):
        show_parent = not self.target.startswith("~")
        target = self.target if show_parent else self.target[1:]
        components = target.split("::")

        try:
            *parents, parent, member = components
            if not parents:
                raise ValueError
        except ValueError:
            msg = self.inliner.reporter.error(
                f"Invalid {self.target_kind} target: {target!r}",
                line=self.lineno,
            )
            prb = self.inliner.problematic(self.rawtext, self.rawtext, msg)

            return [prb], [msg]

        node = reference(
            self.rawtext,
            refuri=(
                f"{BASE_URL}/{'/'.join(parents)}/{self.parent_tag}"
                f".{parent}.html#{self.member_tag}.{member}"
            ),
            **self.options,
        )

        if not self.has_explicit_title:
            title = f"{parent}::{member}" if show_parent else member

        node += literal(
            self.rawtext,
            (
                self.title if self.has_explicit_title
                else f"{title}()" if self.target_is_callable
                else title
            ),
        )

        return [node], []


class InternalsDomain(Domain):
    """A custom Sphinx domain for referencing the librsvg internals docs."""

    name = "internals"
    label = "Librsvg Internals Docs"
    roles = {
        "crate": CrateRole(),
        "module": ModuleRole(),

        # Top-level entities
        "struct": TopLevelRole("struct"),
        "enum": TopLevelRole("enum"),
        "trait": TopLevelRole("trait"),
        "type": TopLevelRole("type"),
        "fn": TopLevelRole("fn", target_is_callable=True),
        "macro": TopLevelRole("macro"),
        "constant": TopLevelRole("constant"),
        "static": TopLevelRole("static"),

        # Member entities
        "struct-field": MemberRole("struct field", "struct", "structfield"),
        "struct-method": MemberRole(
            "struct method", "struct", "method", target_is_callable=True
        ),
        "enum-variant": MemberRole("enum variant", "enum", "variant"),
        "trait-method": MemberRole(
            "provided trait method", "trait", "method", target_is_callable=True
        ),
        "trait-tymethod": MemberRole(
            "required trait method",
            "trait",
            "tymethod",
            target_is_callable=True,
        ),
    }


def setup(app):
    app.add_domain(InternalsDomain)



# Checks that the version of the librsvg public crate matches the version for the GNOME library.
#
# For stable releases:
#   - GNOME: 2.57.2
#   - crate: 2.57.2
#
# For development relases, .9x vs. -beta.x
#   - GNOME: 2.57.90
#   - crate: 2.58.0-beta.0

import semver
import sys
import toml

from utils import get_project_version_str

def gen_crate_version_from_project_version(v):
    if v.patch < 90:
        # stable release, just return it
        return v
    elif v.patch >= 90 and v.patch < 99:
        # development release, mangle it for semver
        patch_level = v.patch - 90
        beta = f'beta.{patch_level}'
        return v.bump_minor().replace(prerelease = beta)
    else:
        raise Exception("don't know what to do with patch versions larger than 99")

def check_crate_version(project_version_str, crate_version_str):
    # GNOME only likes x.y.z versions
    main_version = semver.Version.parse(project_version_str)
    assert main_version.major is not None
    assert main_version.minor is not None
    assert main_version.patch is not None
    assert main_version.prerelease is None
    assert main_version.build is None

    crate_version = semver.Version.parse(crate_version_str)

    if gen_crate_version_from_project_version(main_version) != crate_version:
        raise Exception(
            f'meson.build version {main_version} does not match rsvg crate version {crate_version}'
        )

def test_stable():
    a = semver.Version.parse('2.56.3')
    assert gen_crate_version_from_project_version(a) == a

def test_development():
    a = semver.Version.parse('2.56.90')
    assert gen_crate_version_from_project_version(a) == semver.Version.parse('2.57.0-beta.0')

    a = semver.Version.parse('2.57.93')
    assert gen_crate_version_from_project_version(a) == semver.Version.parse('2.58.0-beta.3')

def main():
    project_version_str = get_project_version_str()

    doc = toml.load('rsvg/Cargo.toml')
    crate_version_str = doc['package']['version']

    check_crate_version(project_version_str, crate_version_str)
    print('Versions number match.  All good!', file=sys.stderr)

if __name__ == "__main__":
    main()



import re

def get_project_version_str():
    regex = re.compile(r" +version: '(\d+\.\d+\.\d+)',")
    with open("meson.build") as f:
        for line in f.readlines():
            matches = regex.match(line)
            if matches is not None:
                version_str = matches.group(1)
                return version_str

    raise Exception('meson.build does not have a version string for the project')



#!/usr/bin/env python3

# Simple script to check whether rustc-version is to be checkedn against minimum supported rust version

import sys
from argparse import ArgumentParser

if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument('--string', help='String to check is a version-like string', required=True)
    args = parser.parse_args()
    parts = args.string.split('.')
    if len(parts) != 2 and len(parts) != 3:
        print('skip')
    else:
        for p in parts:
            try:
                int(p)
            except ValueError:
                print('skip')
                sys.exit()
        print('check')



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import os
import sys
sys.path.insert(0, os.path.abspath('_extensions'))


# -- Project information -----------------------------------------------------

project = 'Development guide for librsvg'
copyright = '2022, Federico Mena Quintero'
author = 'Federico Mena Quintero'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # Used to shorten external links.
    # https://www.sphinx-doc.org/en/master/usage/extensions/extlinks.html
    "sphinx.ext.extlinks",
    # Used to link issues, merge requests, CVEs, etc.
    # https://github.com/sloria/sphinx-issues
    "sphinx_issues",
    # Used to reference entities in the internals documentation.
    # ./_extensions/internals.py
    "internals",
    # Used to reference entries in the source tree.
    # ./_extensions/source.py
    "source",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'furo'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# Options for the linkcheck builder.  This is used by the ci/check_docs_links.sh script.
#
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-the-linkcheck-builder

linkcheck_ignore = [
    # These URLs fail for some reason, but work in the browser.
    r'https://crates.io/crates/.*',

    # Links with anchors for section names or line numbers fail.  But since we usually
    # use specific commit ids, they should be correct anyway.
    r'https://github.com/.*#.*',
    r'https://gitlab.gnome.org/.*#.*',
    r'https://gitlab.freedesktop.org/.*#.*',
]


# Options for the `sphinx.ext.extlinks` extension. See `extensions` above.

extlinks = {
    "rustsec": ("https://rustsec.org/advisories/RUSTSEC-%s", "RUSTSEC-%s"),
}
extlinks_detect_hardcoded_links = True


# Options for the `sphinx-issues` extension. See `extensions` above.

issues_default_group_project = "GNOME/librsvg"
issues_uri = "https://gitlab.gnome.org/{group}/{project}/-/issues/{issue}"
issues_prefix = "#"
issues_pr_uri = "https://gitlab.gnome.org/{group}/{project}/-/merge_requests/{pr}"
issues_pr_prefix = "!"
issues_commit_uri = "https://gitlab.gnome.org/{group}/{project}/-/commit/{commit}"
issues_commit_prefix = "@"
issues_user_uri = "https://gitlab.gnome.org/{user}"
issues_user_prefix = "@"



#!/usr/bin/env python3

from argparse import ArgumentParser
import os
from pathlib import Path
import shutil
import subprocess
import sys

parser = ArgumentParser("Cargo wrapper")

parser.add_argument(
    "--command",
    required=True,
    choices=["cbuild", "test", "build"],
    help="Cargo command",
)

parser.add_argument(
    "--cargo", required=True, type=Path, help="Path to the cargo executable"
)

parser.add_argument(
    "--manifest-path", required=True, type=Path, help="Path to Cargo.toml"
)

parser.add_argument(
    "--current-build-dir",
    required=True,
    type=Path,
    help="Value from meson.current_build_dir()",
)

parser.add_argument(
    "--current-source-dir",
    required=True,
    type=Path,
    help="Value from meson.current_source_dir()",
)

parser.add_argument(
    "--project-build-root",
    required=True,
    type=Path,
    help="Value from meson.project_build_root()",
)

parser.add_argument(
    "--toolchain-version", help="Rust Toolchain Version if needed"
)

parser.add_argument(
    "--target", help="Target triplet"
)

parser.add_argument(
    "--build-triplet", help="Build toolchain triplet (for cross builds using specific toolchain version)"
)

parser.add_argument(
    "--avif", action="store_true", help="Enable AVIF support"
)

parser.add_argument(
    "--pixbuf", action="store_true", help="Enable GDK-Pixbuf support"
)

parser.add_argument(
    "--packages",
    nargs="*",
    default=[],
    help='Rust packages to build (names for "cargo cbuild/build -p")',
)

parser.add_argument(
    "--prefix", type=Path, required=True, help="Value of get_option('prefix')"
)

parser.add_argument("--libdir", required=True, help="Value of get_option('libdir')")

g = parser.add_argument_group("Outputs")
group = parser.add_mutually_exclusive_group(required=False)
group.add_argument(
    "--extension", help="filename extension for the library (so, a, dll, lib, dylib)",
)
group.add_argument("--bin", help="Name of binary to build")

g = parser.add_argument_group("Optimizations")
g.add_argument(
    "--release", action="store_true", help="Build artifacts in release mode"
)
g.add_argument(
    '--optimization', choices=['0', '1', '2', '3', 's'], help="Set optimization level"
)
g.add_argument(
    '--lto', choices=['fat', 'thin'], help="Set optimization level"
)

args = parser.parse_args()

if args.toolchain_version is not None and args.target is None and args.build_triplet is None:
    raise ValueError('--target and/or --build-triplet argument required if --toolchain-version is specified')

if args.command == 'test':
    if args.extension or args.bin:
        raise ValueError('Cargo test does not take --extension or --bin')

cargo_target_dir = Path(args.project_build_root) / "target"

# The final rsvg-convert executable will be found in cargo_target_dir/$(TARGET_TRIPLET)
# if a target triplet is specified
if args.target:
    cargo_target_output_dir = cargo_target_dir / args.target
else:
    cargo_target_output_dir = cargo_target_dir

env = os.environ.copy()
pkg_config_path = [i for i in env.get("PKG_CONFIG_PATH", "").split(os.pathsep) if i]
pkg_config_path.insert(
    0, (Path(args.project_build_root) / "meson-uninstalled").as_posix()
)
env["PKG_CONFIG_PATH"] = os.pathsep.join(pkg_config_path)

features = []

if args.avif:
    features.append('avif')

if args.pixbuf:
    features.append('pixbuf')

cargo_prefixes = [
    "--prefix",
    Path(args.prefix).as_posix(),
    "--libdir",
    (Path(args.prefix) / args.libdir).as_posix(),
]

cargo_cmd = [Path(args.cargo).as_posix()]

if args.toolchain_version is not None:
    if args.build_triplet is not None:
        cargo_cmd.extend(["+%s-%s" % (args.toolchain_version, args.build_triplet)])
    else:
        cargo_cmd.extend(["+%s-%s" % (args.toolchain_version, args.target)])

if args.command == "cbuild":
    cargo_cmd.extend(["cbuild", "--locked"])
    library_type = "staticlib" if args.extension in ("a", "lib") else "cdylib"
    cargo_cmd.extend(cargo_prefixes)
    cargo_cmd.extend(["--library-type", library_type])
elif args.command == "test":
    cargo_cmd.extend(["test", "--locked", "--no-fail-fast", "--color=always"])
    if 'librsvg' in args.packages:
        # These are required for librsvg itself
        # If doing an unqualified cargo build, they'll be called up
        # by rsvg-convert
        # https://github.com/rust-lang/cargo/issues/2911
        features.extend(["capi", "test-utils"])
else:
    cargo_cmd.extend(["build", "--locked"])
    if args.bin:
        cargo_cmd.extend(["--bin", args.bin])

cargo_cmd.extend(["--manifest-path", Path(args.manifest_path).as_posix()])
cargo_cmd.extend(["--target-dir", cargo_target_dir.as_posix()])

if args.release:
    buildtype = 'release'
    cargo_cmd.extend(['--release'])
else:
    buildtype = 'debug'

if args.optimization:
    env[f'CARGO_PROFILE_{buildtype.upper()}_OPT_LEVEL'] = args.optimization
if args.lto:
    env[f'CARGO_PROFILE_{buildtype.upper()}_CODEGEN_UNITS'] = '1'
    env[f'CARGO_PROFILE_{buildtype.upper()}_LTO'] = args.lto

if args.target:
    cargo_cmd.extend(['--target', args.target])

if features:
    cargo_cmd.extend(["--features", ",".join(features)])

for p in args.packages:
    cargo_cmd.extend(["-p", p])

if args.command == "test":
    cargo_cmd.extend(["--", "--include-ignored"])

k = {k: v for k, v in env.items() if k.startswith('CARGO_PROFILE')}
print(f"command: {cargo_cmd}, env: {k}")
subprocess.run(cargo_cmd, env=env, check=True)

if args.command in ["cbuild", "build"]:
    # Copy so/dll/etc files to build dir
    if args.extension:
        for f in cargo_target_dir.glob(f"**/{buildtype}/*.{args.extension}"):
            shutil.copy(f, args.current_build_dir)
    # Copy binary and, if applicable, the corresponding .pdb file, to build dir
    else:
        binary = Path(cargo_target_output_dir / buildtype / args.bin)
        is_windows_target = False
        if args.target is None and sys.platform == "win32":
            is_windows_target = True
        elif args.target is not None and args.target.split('-')[2] == 'windows':
            is_windows_target = True
        if is_windows_target:
            exe_name = args.bin.replace('rsvg-convert', 'rsvg_convert')
            pdb_src = Path(cargo_target_output_dir / buildtype / exe_name).with_suffix(".pdb")
            pdb_dest = Path(args.current_build_dir / args.bin).with_suffix('.pdb')
            if pdb_src.exists():
                shutil.copy(pdb_src, pdb_dest)
            binary = binary.with_suffix(".exe")
        shutil.copy(binary, args.current_build_dir)



"""Custom reStructuredText entities/helpers for referencing entries in the
librsvg source tree.

For any changes made in this module, please ensure
``devel-docs/devel_docs_mod_guide.rst`` is updated accordingly, if necessary.

For help/reference on modifying these, see:

- https://www.sphinx-doc.org/en/master/development/tutorials/extending_syntax.html
- https://docutils.sourceforge.io/docs/howto/rst-roles.html
- https://protips.readthedocs.io/link-roles.html
- https://www.sphinx-doc.org/en/master/extdev/utils.html#sphinx.util.docutils.SphinxRole
- https://www.sphinx-doc.org/en/master/extdev/utils.html#sphinx.util.docutils.ReferenceRole
- https://github.com/sphinx-doc/sphinx/blob/master/sphinx/roles.py
"""

from __future__ import annotations

from docutils.nodes import reference
from sphinx.util.docutils import ReferenceRole

BASE_URL = "https://gitlab.gnome.org/GNOME/librsvg/-/tree"


class SourceRole(ReferenceRole):
    """A custom Sphinx role for referencing entries in the librsvg source tree.
    """

    def run(self):
        ref, _, path = self.target.rpartition(":")

        if path.startswith("/"):
            msg = self.inliner.reporter.error(
                f"Invalid source tree entry path: {path!r}", line=self.lineno
            )
            prb = self.inliner.problematic(self.rawtext, self.rawtext, msg)

            return [prb], [msg]

        if ref:
            # 12 <= 9 characters + elipsis (3 periods);
            # 12, to allow `librsvg-X.YZ`.
            short_ref = f"{ref[:9]}..." if len(ref) > 12 else ref
        else:
            short_ref = ref = "main"

        node = reference(
            self.rawtext,
            (
                self.title if self.has_explicit_title
                else path or 'the source tree' if ref == "main"
                else f"{path or 'the source tree'} (@ {short_ref})"
            ),
            refuri=f"{BASE_URL}/{ref}/{path}",
            **self.options,
        )

        return [node], []


def setup(app):
    app.add_role("source", SourceRole())



#!/usr/bin/env python3

from argparse import ArgumentParser
from pathlib import Path
import os
import subprocess
import sys

argparse = ArgumentParser('Deploy loaders.cache')

argparse.add_argument('--gdk-pixbuf-moduledir', type=Path)
argparse.add_argument('gdk_pixbuf_queryloaders', type=Path)
argparse.add_argument('gdk_pixbuf_cache_file', type=Path)
argparse.add_argument('--show-cross-message', action='store_true')

if __name__ == '__main__':
    args = argparse.parse_args()

    cache_file: Path = args.gdk_pixbuf_cache_file

    if args.show_cross_message or os.environ.get("DESTDIR"):
        print('*** Note: Please run gdk-pixbuf-queryloaders manually ' +
              'against the newly-built gdkpixbuf-svg loader', file=sys.stderr)
    else:
        env = os.environ.copy()
        env['GDK_PIXBUF_MODULEDIR'] = Path(args.gdk_pixbuf_moduledir).as_posix()
        with cache_file.open('w', encoding='utf-8') as f:
            subprocess.run(
                [Path(args.gdk_pixbuf_queryloaders).as_posix()],
                env=env,
                stdout=f,
                check=True
            )




# This script checks that the project's version is the same in a few files where it must appear.

import sys
import toml

from utils import get_project_version_str

def get_cargo_toml_version():
    doc = toml.load('Cargo.toml')
    return doc['workspace']['package']['version']

def get_doc_version():
    doc = toml.load('doc/librsvg.toml')
    return doc['library']['version']

def main():
    versions = [
        ['meson.build', get_project_version_str()],
        ['Cargo.toml', get_cargo_toml_version()],
        ['doc/librsvg.toml', get_doc_version()],
    ]

    all_the_same = True

    for filename, version in versions[1:]:
        if version != versions[0][1]:
            all_the_same = False

    if not all_the_same:
        print('Version numbers do not match, please fix them!\n', file=sys.stderr)
        for filename, version in versions:
            print(f'{filename}: {version}', file=sys.stderr)

        sys.exit(1)

    print('Versions number match.  All good!', file=sys.stderr)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) 2022 L. E. Segovia <amy@amyspark.me>
#
# This file is part of the FFmpeg Meson build
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http://www.gnu.org/licenses/>.

import argparse
import errno
import os
import pathlib
import re
import subprocess


def output(platform, symbols):
    if platform == 'win':
        print("EXPORTS")
        print(*[f'    {symbol}' for symbol in sorted(set(symbols))], sep='\n')
    elif platform == 'darwin':
        print(*[f'{prefix}{symbol}' for symbol in sorted(set(symbols))], sep='\n')
    else:
        print('{')
        print('    global:')
        print(
            *[f'        {prefix}{symbol};' for symbol in sorted(set(symbols))], sep='\n')
        print('    local:')
        print('        *;')
        print('};')


if __name__ == '__main__':
    arg_parser = argparse.ArgumentParser(
        description='Craft the symbols exports file')

    arg_parser.add_argument('--prefix', metavar='PREFIX',
                            help='Prefix for extern symbols')
    g = arg_parser.add_argument_group('Library parsing tool')
    group = g.add_mutually_exclusive_group(required=True)
    group.add_argument('--nm', metavar='NM_PATH', type=pathlib.Path,
                       help='If specified, runs this instead of dumpbin (MinGW)')
    group.add_argument('--dumpbin', metavar='DUMPBIN_PATH', type=pathlib.Path,
                       help='If specified, runs this instead of nm (MSVC)')
    group.add_argument(
        '--list', action='store_true', help='If specified, consider FILE as an exported symbols list instead of a library')
    g = arg_parser.add_argument_group('Symbol naming')
    group = g.add_mutually_exclusive_group(required=True)
    group.add_argument('--regex', metavar='REGEX', type=str,
                       nargs='+',
                       help='Regular expression for exported symbols')
    group.add_argument('--vscript', metavar='VERSION_SCRIPT',
                       type=argparse.FileType('r'), help='Version script')
    arg_parser.add_argument('--os', type=str, choices=('win', 'linux', 'darwin'),
                            default='linux', required=True,
                            help='Target operating system for the exports file (win = MSVC module definition file, linux = version script, darwin = exported symbols list)')
    arg_parser.add_argument('libnames', metavar='FILE', type=pathlib.Path,
                            nargs='+',
                            help='Source file(s) to parse')

    args = arg_parser.parse_args()

    libnames = args.libnames

    for libname in libnames:
        if not libname.exists():
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), libname)

    if not args.list and len(libnames) > 1:
        raise ValueError("Expect 1 filename as argument.")

    prefix = args.prefix or ''
    started = 0
    regex = []

    if args.vscript:
        for line in args.vscript:
            # We only care about global symbols
            if re.match(r'^\s+global:', line):
                started = 1
                line = re.sub(r'^\s+global: *', '', line)
            else:
                if re.match(r'^\s+local:', line):
                    started = 0

            if started == 0:
                continue

            line = line.replace(';', '')

            for exp in line.split():
                # Remove leading and trailing whitespace
                regex.append(exp.strip())
    else:
        regex.extend(args.regex)

    # Ensure things are compatible on Windows with Python 3.7.x
    libname_path_posix = pathlib.Path(libnames[0]).as_posix()

    if args.list:
        dump = []
        for libname in libnames:
            syms = libname.open('r', encoding='utf-8').readlines()
            # Strip whitespaces
            syms = [x.strip() for x in syms]
            # Exclude blank lines
            syms = [x for x in syms if len(x) > 0]
            dump.extend(syms)
    elif args.nm is not None:
        # Use eval, since NM="nm -g"
        # Add -j to ensure only symbol names are output (otherwise in macOS
        # a race condition can occur in the redirection)
        # And use `--no-llvm-bc` in case it's /usr/bin/nm on macOS
        s = subprocess.run(
            [args.nm, '-U', '-g', '-j', '--no-llvm-bc', libname_path_posix],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            check=False,
        )
        if s.returncode != 0:
            # If it fails, retry with --defined-only (non macOS)
            s = subprocess.run(
                [args.nm, '--defined-only', '-g', '-j', '--no-llvm-bc', libname_path_posix],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                check=False,
            )
        if s.returncode != 0:
            # If it fails, retry without skipping LLVM bitcode (macOS flag)
            # Don't use -U, as that was an alias for --unicode= instead of
            # --defined-only before Binutils 2.39
            s = subprocess.run(
                [args.nm, '--defined-only', '-g', '-j', libname_path_posix],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                check=False,
            )
        if s.returncode != 0:
            # -j was added only in Binutils 2.37
            s = subprocess.run(
                [args.nm, '--defined-only', '-g', libname_path_posix],
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
                universal_newlines=True,
                check=True,
            )
        dump = s.stdout.splitlines()
        # Exclude lines with ':' (object name)
        dump = [x for x in dump if ":" not in x]
        # Exclude blank lines
        dump = [x for x in dump if len(x) > 0]
        # Subst the prefix out
        dump = [re.sub(f'^{prefix}', '', x) for x in dump]
    else:
        dump = subprocess.run([pathlib.Path(args.dumpbin).as_posix(), '-linkermember:1', libname_path_posix],
                              stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True).stdout.splitlines()
        # Find the index of the first line with
        # "public symbols", keep the rest
        # Then the line with " Summary",
        # delete it and the rest
        for i, line in enumerate(dump):
            if 'public symbols' in line:
                start = i
            elif re.match(r'\s+Summary', line):
                end = i
        dump = dump[start:end]
        # Substitute prefix out
        dump = [re.sub(fr'\s+{prefix}', ' ', x) for x in dump]
        # Substitute big chonky spaces out
        dump = [re.sub(r'\s+', ' ', x) for x in dump]
        # Exclude blank lines
        dump = [x for x in dump if len(x) > 0]
        # Take only the *second* field (split by spaces)
        # Python's split excludes whitespace at the beginning
        dump = [x.split()[1] for x in dump]

    symbols = []
    for exp in regex:
        for i in dump:
            if re.match(exp, i):
                symbols.append(i)

    output(args.os, symbols)
